/*
 * sar: report system activity
 * (C) 1999 by Sebastien GODARD (sebastien.godard@wanadoo.fr)
 * 
 ***************************************************************************
 * This program is free software; you can redistribute it and/or modify it *
 * under the terms of the GNU General Public License as published  by  the *
 * Free Software Foundation; either version 2 of the License, or (at  your *
 * option) any later version.                                              *
 *                                                                         *
 * This program is distributed in the hope that it  will  be  useful,  but *
 * WITHOUT ANY WARRANTY; without the implied warranty  of  MERCHANTABILITY *
 * or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License *
 * for more details.                                                       *
 *                                                                         *
 * You should have received a copy of the GNU General Public License along *
 * with this program; if not, write to the Free Software Foundation, Inc., *
 * 675 Mass Ave, Cambridge, MA 02139, USA.                                 *
 ***************************************************************************
 */

/* Version */
#define VERSION		"1.1"

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>
#include <sys/utsname.h>
#include <sys/stat.h>
#include <linux/kernel_stat.h>   /* For NR_IRQS */
#include "sar.h"

#ifdef USE_NLS

#include <libintl.h>
#define _(string) gettext(string)

#else

#define _(string) (string)

#endif


int actflag = 0;
long interval = 0;
long int count = 1;
long file_interval;
/* 32nd bit of int_stat[NR_IRQS / 4] reserved for total nb of interrupts */
unsigned int int_stat[(NR_IRQS / 32) + 1];
/* 32nd bit of cpu_stat[(NR_CPUS / 32) + 1] reserved for global cpu */
unsigned int cpu_stat[(NR_CPUS / 32) + 1];
char stat_line[MAX_LINE_LEN];
char buffer[BUF_SIZE];
short int sl_hour, sl_minute, sl_second;
unsigned long file_uptime = 0;
char tm_start[9], tm_end[9];

/*
 * array[0]: stat at t
 * array[1]: stat at t'
 * array[2]: used to compute the average.
 */
unsigned int cpu_user[3] = {0, 0, 0}, cpu_nice[3] = {0, 0, 0}, cpu_system[3] = {0, 0, 0};
unsigned long cpu_idle[3] = {0, 0, 0};
unsigned int per_cpu_user[3][NR_CPUS], per_cpu_nice[3][NR_CPUS], per_cpu_system[3][NR_CPUS];
unsigned long per_cpu_idle[3][NR_CPUS];
unsigned int dk_drive[3] = {0, 0, 0};
unsigned int u_tmp[NR_DISK - 1];
unsigned int dk_drive_rio[3] = {0, 0, 0}, dk_drive_wio[3] = {0, 0, 0};
unsigned int dk_drive_rblk[3] = {0, 0, 0}, dk_drive_wblk[3] = {0, 0, 0};
unsigned long uptime_h, idle_h, uptime[3] = {0, 0, 0};
unsigned int uptime_l, idle_l;
unsigned int pgpgin[3] = {0, 0, 0}, pgpgout[3] = {0, 0, 0};
unsigned int pswpin[3] = {0, 0, 0}, pswpout[3] = {0, 0, 0};
unsigned int interrupts[3][NR_IRQS];
unsigned int int_sum[3] = {0, 0, 0};
unsigned int context_swtch[3] = {0, 0, 0};
unsigned long total_forks[3] = {0, 0, 0};


/*
 * Print usage and exit
 */
void usage(char *progname)
{
   fprintf(stderr, _("%s version %s\n"
	           "Usage: %s [options...]\n"
	           "Options are:\n"
	           "   [-A] [-c] [-g] [-i] [-r] [-u] [-w] [-W]\n"
		   "   [-I { irq | -1 } ] [-U { cpu | -1 } ]\n"
	           "   [ {-o [filename] | -f [filename] [-s [hh:mm:ss] ] } [-e [hh:mm:ss] ] ]\n"
	           "   [interval [count] ]\n"),
	   progname, VERSION, progname);
   exit(1);
}


/* 
 * SIGALRM signal handler
 */
void alarm_handler(int sig)
{
   signal(SIGALRM, alarm_handler);
   alarm(interval);
}


/*
 * Init irqs array
 */
void init_int_stat(unsigned int value)
{
   int i;
   
   for (i = 0; i <= NR_IRQS >> 5; i++)
     int_stat[i] = value;
}


/*
 * Init CPUs array
 */
void init_cpu_stat(unsigned int value)
{
   int i;
   
   for (i = 0; i <= NR_CPUS >> 5; i++)
     cpu_stat[i] = value;
}


/*
 * Get current date or time
 */
char *get_date_or_time(int type)
{
   time_t timer;
   struct tm *loc_time;
   static char buffer[11]; 
   
   time(&timer);
   loc_time = localtime(&timer);
   if (type == DT_TIME)
      /* Get current time */
      sprintf(buffer, "%02d:%02d:%02d",
	      loc_time->tm_hour, loc_time->tm_min, loc_time->tm_sec);
   else if (type == DT_DATE)
      /* Get current date */
      sprintf(buffer, "%4d/%02d/%02d",
	      loc_time->tm_year + 1900, loc_time->tm_mon + 1, loc_time->tm_mday);
   else
      /* Get current day */
      sprintf(buffer, "%02d", loc_time->tm_mday);
   
   return buffer;
}


/*
 * Next time stamp
 */
int next_slice(void)
{
   int min, max, r, pt1, pt2;
   

   sl_second += file_interval;
   if (sl_second >= 60) {
      sl_minute += sl_second / 60;
      sl_second %= 60;
      if (sl_minute >= 60) {
	 sl_hour += sl_minute / 60;
	 sl_minute %= 60;
	 if (sl_hour >= 24)
	    sl_hour %= 24;
      }
   }

   /* Display all entries if interval is -1 */
   if (interval == -1)
      return 1;

   /*
    * A few notes about the algorithm used here to display selected entries
    * from the system activity file (option -f with a given interval number):
    * Let Iu be the interval value given by the user on the command line,
    *     If the interval value used in the system activity file,
    * and En the nth entry (identified by its time stamp) of the file.
    * We choose In = [ En - If/2, En + If/2 [ if If is even,
    *        or In = [ En - If/2, En + If/2 ] if not.
    * En will be displayed if
    *       (Pn * Iu) /or/ (P'n * Iu) belongs to In
    * with  Pn = En / Iu and P'n = En / Iu + 1
    */
   min = (file_uptime / 100) - (file_interval / 2);
   max = (file_uptime / 100) + (file_interval / 2) + (file_interval % 2);
   
   pt1 = ((file_uptime / 100) / interval) * interval;
   pt2 = (((file_uptime / 100) / interval) + 1) * interval;
   
   r   = (((pt1 >= min) && (pt1 < max)) || ((pt2 >= min) && (pt2 < max)));

#ifdef DEBUG
   printf("Sec: %02d  Min: %d  Max: %d  En: %ld  Pn: %ld  P'n: %ld  Pt1: %d  Pt2: %d  r: %d\n",
	  sl_second, min, max, file_uptime / 100,
	  (file_uptime / 100) / interval, (file_uptime / 100) / interval + 1,
	  pt1, pt2, r);
#endif   
   return r;
}


/*
 * Initialize interrupt stat counters
 */
void init_stat_irq(void)
{
   int nr_irq, i;
   
   for (i = 0; i < 3; i++)
      for (nr_irq = 0; nr_irq < NR_IRQS; nr_irq++)
	 interrupts[i][nr_irq] = 0;
}


/*
 * Initialize cpu stat counters
 */
void init_stat_cpu(void)
{
   int nr_cpu, i;
   
   for (i = 0; i < 3; i++)
      for (nr_cpu = 0; nr_cpu < NR_CPUS; nr_cpu++) {
	 per_cpu_user[i][nr_cpu] = 0;
	 per_cpu_nice[i][nr_cpu] = 0;
	 per_cpu_system[i][nr_cpu] = 0;
	 per_cpu_idle[i][nr_cpu] = 0;
      }
}


/*
 * Open system activity files for reading and/or writing
 */
void open_files(char from_file[], char to_file[], int *from_fd, int *to_fd)
{

   if (!from_file[0] && !to_file[0])
      return;
   
   /* Check file names */
   if (!strcmp(from_file, to_file)) {
      fprintf(stderr, _("Input and output files are identical\n"));
      exit(2);
   }
   
   if (from_file[0]) {
      /* Open input daily file */
      if ((*from_fd = open(from_file, O_RDONLY)) < 0) {
	 fprintf(stderr, _("Cannot open %s: %s\n"), from_file, strerror(errno));
	 exit(3);
      }
   }
   
   if (to_file[0]) {
      /* Open output daily file */
      if ((*to_fd = open(to_file, O_WRONLY | O_CREAT | O_TRUNC,
			 S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH)) < 0) {
	 fprintf(stderr, _("Cannot open %s: %s\n"), to_file, strerror(errno));
	 exit(3);
      }
   }
}


/*
 * Close system activity files
 */
void close_files(char from_file[], char to_file[], int from_fd, int to_fd)
{
   if (from_file[0])
      close(from_fd);
   if(to_file[0])
      close(to_fd);
}


/*
 ***********************************
 * Read machine uptime
 ***********************************
 */
void read_uptime(int curr)
{
   FILE *uptimefp;
   char line[80];


   /* Open uptime file */
   if ((uptimefp = fopen(UPTIME, "r")) == NULL) {
      fprintf(stderr, _("Cannot open %s: %s\n"), UPTIME, strerror(errno));
      exit(1);
   }
   
   if (fgets(line, 80, uptimefp) == NULL) {
      fprintf(stderr, _("Cannot read %s file\n"), UPTIME);
      exit(1);
   }

   /*
    * Read the uptime of the system, and the amount of time spent in the idle process
    * (expressed in seconds)
    */
   sscanf(line, "%lu.%u %lu.%u", &uptime_h, &uptime_l, &idle_h, &idle_l);

   /* Close stat file */
   fclose(uptimefp);
   
   /* Compute the uptime of the system in jiffies (1/100ths of a second) */
   uptime[curr] = uptime_h * 100 + uptime_l;
}


/*
 * Save stats values (will be used to compute stats average)
 */
void trans_stat(int save, int curr)
{
   int i;
   
   
   total_forks[save] = total_forks[curr];
   context_swtch[save] = context_swtch[curr];
   cpu_user[save] = cpu_user[curr];
   cpu_nice[save] = cpu_nice[curr];
   cpu_system[save] = cpu_system[curr];
   cpu_idle[save] = cpu_idle[curr];
   
   int_sum[save] = int_sum[curr];
   for (i = 0; i < NR_IRQS; i++)
      interrupts[save][i] = interrupts[curr][i];
   
   pgpgin[save] = pgpgin[curr];
   pgpgout[save] = pgpgout[curr];
   pswpin[save] = pswpin[curr];
   pswpout[save] = pswpout[curr];
   dk_drive[save] = dk_drive[curr];
   dk_drive_rio[save] = dk_drive_rio[curr];
   dk_drive_wio[save] = dk_drive_wio[curr];
   dk_drive_rblk[save] = dk_drive_rblk[curr];
   dk_drive_wblk[save] = dk_drive_wblk[curr];
   
   uptime[save] = uptime[curr];
}


/*
 ***********************************
 * Read stats from system activity file
 ***********************************
 */
int read_stat_from_file(int curr, int from_fd)
{
   int nb, pos = 0;
   
   
   nb = read(from_fd, buffer, ENTRY_SIZE);
   if (!nb)
      /* End of system activity file */
      return 1;
   if (nb < 0) {
      fprintf(stderr, _("Error while reading system activity file: %s\n"), strerror(errno));
      exit(4);
   }
   else if (nb < ENTRY_SIZE) {
      fprintf(stderr, _("End of system activity file unexpected\n"));
      exit(4);
   }
   
   memcpy(&total_forks[curr], &buffer[pos], sizeof(long));
   pos += sizeof(long);
   memcpy(&context_swtch[curr], &buffer[pos], sizeof(int));
   pos += sizeof(int);
   memcpy(&cpu_user[curr], &buffer[pos], sizeof(int));
   pos += sizeof(int);
   memcpy(&cpu_nice[curr], &buffer[pos], sizeof(int));
   pos += sizeof(int);
   memcpy(&cpu_system[curr], &buffer[pos], sizeof(int));
   pos += sizeof(int);
   memcpy(&cpu_idle[curr], &buffer[pos], sizeof(long));
   pos += sizeof(long);
   memcpy(&int_sum[curr], &buffer[pos], sizeof(int));
   pos += sizeof(int);

   for (nb = 0; nb < NR_IRQS; nb++) {
      memcpy(&interrupts[curr][nb], &buffer[pos], sizeof(int));
      pos += sizeof(int);
   }
   
   memcpy(&pgpgin[curr], &buffer[pos], sizeof(int));
   pos += sizeof(int);
   memcpy(&pgpgout[curr], &buffer[pos], sizeof(int));
   pos += sizeof(int);
   memcpy(&pswpin[curr], &buffer[pos], sizeof(int));
   pos += sizeof(int);
   memcpy(&pswpout[curr], &buffer[pos], sizeof(int));
   pos += sizeof(int);
   memcpy(&dk_drive[curr], &buffer[pos], sizeof(int));
   pos += sizeof(int);
   memcpy(&dk_drive_rio[curr], &buffer[pos], sizeof(int));
   pos += sizeof(int);
   memcpy(&dk_drive_wio[curr], &buffer[pos], sizeof(int));
   pos += sizeof(int);
   memcpy(&dk_drive_rblk[curr], &buffer[pos], sizeof(int));
   pos += sizeof(int);
   memcpy(&dk_drive_wblk[curr], &buffer[pos], sizeof(int));
   pos += sizeof(int);
   
   return 0;
}


/*
 ***********************************
 * Read stats...
 * (see linux source file linux/fs/proc/array.c)
 ***********************************
 */
void read_stat(int curr)
{
   FILE *statfp;
   char resource_name[16];
   char line[2240];
   int i, pos;
   
   
   /* Open stat file */
   if ((statfp = fopen(STAT, "r")) == NULL) {
      fprintf(stderr, _("Cannot open %s: %s\n"), STAT, strerror(errno));
      exit(1);
   }

   while (fgets(line, 2240, statfp) != NULL) {
      
      if (!strncmp(line, "cpu ", 4)) {
	 /* Read the number of jiffies spent in user, nice, system and idle mode */
	 sscanf(line, "%s  %u %u %u %lu",
	        resource_name, &cpu_user[curr], &cpu_nice[curr],
		&cpu_system[curr], &cpu_idle[curr]);
      }

      else if (!strncmp(line, "disk ", 5)) {
	 /* Read number of I/O done since the last reboot */
	 sscanf(line, "%s %u %u %u %u",
		resource_name, &dk_drive[curr], &u_tmp[0], &u_tmp[1], &u_tmp[2]);
	 dk_drive[curr] += u_tmp[0] + u_tmp[1] + u_tmp[2];
      }
   
      else if (!strncmp(line, "disk_rio", 8)) {
	 /* Read number of read I/O */
	 sscanf(line, "%s %u %u %u %u",
		resource_name, &dk_drive_rio[curr], &u_tmp[0], &u_tmp[1], &u_tmp[2]);
	 dk_drive_rio[curr] += u_tmp[0] + u_tmp[1] + u_tmp[2];
      }

      else if (!strncmp(line, "disk_wio", 8)) {
	 /* Read number of write I/O */
	 sscanf(line, "%s %u %u %u %u",
		resource_name, &dk_drive_wio[curr], &u_tmp[0], &u_tmp[1], &u_tmp[2]);
	 dk_drive_wio[curr] += u_tmp[0] + u_tmp[1] + u_tmp[2];
      }

      else if (!strncmp(line, "disk_rblk", 9)) {
	 /* Read number of blocks read from disk (one block is 1024 bytes) */
	 sscanf(line, "%s %u %u %u %u",
		resource_name, &dk_drive_rblk[curr], &u_tmp[0], &u_tmp[1], &u_tmp[2]);
	 dk_drive_rblk[curr] += u_tmp[0] + u_tmp[1] + u_tmp[2];
      }

      else if (!strncmp(line, "disk_wblk", 9)) {
	 /* Read number of blocks written to disk */
	 sscanf(line, "%s %u %u %u %u",
		resource_name, &dk_drive_wblk[curr], &u_tmp[0], &u_tmp[1], &u_tmp[2]);
	 dk_drive_wblk[curr] += u_tmp[0] + u_tmp[1] + u_tmp[2];
      }
      
      else if (!strncmp(line, "page", 4)) {
	 /* Read number of pages the system paged in and out */
	 sscanf(line, "%s %u %u",
		resource_name, &pgpgin[curr], &pgpgout[curr]);
      }
      
      else if (!strncmp(line, "swap", 4)) {
	 /* Read number of swap pages brought in and out */
	 sscanf(line, "%s %u %u",
		resource_name, &pswpin[curr], &pswpout[curr]);
      }

      else if (!strncmp(line, "intr", 4)) {
	 /* Read total number of interrupts received since system boot */
	 sscanf(line, "%s %u", resource_name, &int_sum[curr]);
	 pos = strcspn(line + 5, " ") + 5;
	 
	 /* Read number of each interrupts received since system boot */
	 for (i = 0; i < NR_IRQS; i++) {
	    sscanf(line + pos, " %u", &interrupts[curr][i]);
	    pos += strcspn(line + pos + 1, " ") + 1;
	 }
      }
	 
#if 0      
      else if (!strncmp(line, "intr", 4)) {
	 /* Read number of interrupts received since system boot */
	 sscanf(line, "%s %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u %u",
		resource_name, &int_sum[curr],
		&interrupts[curr][0], &interrupts[curr][1], &interrupts[curr][2],
		&interrupts[curr][3], &interrupts[curr][4], &interrupts[curr][5],
		&interrupts[curr][6], &interrupts[curr][7], &interrupts[curr][8],
		&interrupts[curr][9], &interrupts[curr][10], &interrupts[curr][11],
		&interrupts[curr][12], &interrupts[curr][13], &interrupts[curr][14],
		&interrupts[curr][15]);
      }
#endif
      
      else if (!strncmp(line, "ctxt", 4)) {
	 /* Read number of context switches */
	 sscanf(line, "%s %u",
		resource_name, &context_swtch[curr]);
      }
      
      else if (!strncmp(line, "processes", 9)) {
	 /* Read number of processes created since system boot */
	 sscanf(line, "%s %lu",
		resource_name, &total_forks[curr]);
      }
   }
   /* Close stat file */
   fclose(statfp);
}


/*
 ***********************************
 * Print report header
 ***********************************
 */
void print_report_hdr(int read_from_file, int from_fd, int write_to_file, int to_fd)
{
   struct utsname header;
   char banner[MAX_LINE_LEN];
   int len, nb, i = 0;
   unsigned int mask;
   char cur_date[11], cur_time[9];
   char cv[3];

   
   if (read_from_file) {
      
      /* Read system activity file header */
      nb = read(from_fd, buffer, HEADER_SIZE);
      if ((nb != HEADER_SIZE) || ((buffer[0] + buffer[1] * 256) != SA_MAGIC)) {
	 fprintf(stderr, _("Invalid system activity file\n"));
	 exit(4);
      }
      /* Get date */
      strncpy(cur_date, &buffer[2], 10);
      cur_date[10] = '\0';
      /* Get time */
      strncpy(cur_time, &buffer[12], 8);
      cur_time[8] = '\0';
      /* Get interval */
      memcpy(&file_interval, &buffer[20], sizeof(long));
      /* Set time variables */
      cv[2] = '\0';
      sl_hour = atol(strncpy(cv, cur_time, 2));
      sl_minute = atol(strncpy(cv, &cur_time[3], 2));
      sl_second = atol(strncpy(cv, &cur_time[6], 2));
   }
   else {
      strcpy(cur_date, get_date_or_time(DT_DATE));
      strcpy(cur_time, get_date_or_time(DT_TIME));
   }


   if (write_to_file) {

      /* Write header file */
      buffer[0] = SA_MAGIC & 0x00ff;	/* System activity file magic number */
      buffer[1] = SA_MAGIC >> 8;
      sprintf(&buffer[2], "%s", cur_date);	/* Date */
      sprintf(&buffer[12], "%s", cur_time);	/* Time */
      if (interval < 0)
	 memcpy(&buffer[20], &file_interval, sizeof(long));
      else
	 memcpy(&buffer[20], &interval, sizeof(long));
      
      if ((nb = write(to_fd, buffer, HEADER_SIZE)) != HEADER_SIZE) {
	 fprintf(stderr, _("Cannot write system activity file header: %s\n"), strerror(errno));
	 exit(3);
      }
   }

   /* Get system info */
   uname(&header);
   
   printf("%s %s (%s) \t%s\n",
	  header.sysname, header.release, header.nodename, cur_date);

   sprintf(banner, "%s  ", cur_time);
   if (GET_FORK(actflag))
      strcat(banner, "    fork/s");
   if (GET_CTXSW(actflag))
      strcat(banner, "   cswch/s");
   if (GET_CPU(actflag))
      strcat(banner, "     %user     %nice   %system     %idle");
   if (GET_INT(actflag)) {
      if (int_stat[NR_IRQS >> 5] & 0x80000000)
         strcat(banner, "    intr/s");
      for (i = 0; i < NR_IRQS; i++) {
	 mask = 1 << (i & 0x1f);
	 if (int_stat[i >> 5] & mask) {
	    len = strlen(banner);
	    sprintf(banner + len, "  irq%03d/s", i);
	 }
      }
   }
   if (GET_PAGE(actflag))
      strcat(banner, "  pgpgin/s pgpgout/s");
   if (GET_SWAP(actflag))
      strcat(banner, "  pswpin/s pswpout/s");
   if (GET_IO(actflag))
      strcat(banner, "       tps      rtps      wtps Kb_read/s Kb_wrtn/s");
   printf("\n%s\n", banner);
}


/*
 ***********************************
 * Write *everything* to system activity daily file
 ***********************************
 */
void write_stat(int curr, int to_fd)
{
   int pos = 0;
   int i;
   
   
   memcpy(&buffer[pos], &total_forks[curr], sizeof(long));
   pos += sizeof(long);
   memcpy(&buffer[pos], &context_swtch[curr], sizeof(int));
   pos += sizeof(int);
   memcpy(&buffer[pos], &cpu_user[curr], sizeof(int));
   pos += sizeof(int);
   memcpy(&buffer[pos], &cpu_nice[curr], sizeof(int));
   pos += sizeof(int);
   memcpy(&buffer[pos], &cpu_system[curr], sizeof(int));
   pos += sizeof(int);
   memcpy(&buffer[pos], &cpu_idle[curr], sizeof(long));
   pos += sizeof(long);
   memcpy(&buffer[pos], &int_sum[curr], sizeof(int));
   pos += sizeof(int);

   for (i = 0; i < NR_IRQS; i++) {
      memcpy(&buffer[pos], &interrupts[curr][i], sizeof(int));
      pos += sizeof(int);
   }
   
   memcpy(&buffer[pos], &pgpgin[curr], sizeof(int));
   pos += sizeof(int);
   memcpy(&buffer[pos], &pgpgout[curr], sizeof(int));
   pos += sizeof(int);
   memcpy(&buffer[pos], &pswpin[curr], sizeof(int));
   pos += sizeof(int);
   memcpy(&buffer[pos], &pswpout[curr], sizeof(int));
   pos += sizeof(int);
   memcpy(&buffer[pos], &dk_drive[curr], sizeof(int));
   pos += sizeof(int);
   memcpy(&buffer[pos], &dk_drive_rio[curr], sizeof(int));
   pos += sizeof(int);
   memcpy(&buffer[pos], &dk_drive_wio[curr], sizeof(int));
   pos += sizeof(int);
   memcpy(&buffer[pos], &dk_drive_rblk[curr], sizeof(int));
   pos += sizeof(int);
   memcpy(&buffer[pos], &dk_drive_wblk[curr], sizeof(int));
   pos += sizeof(int);
   
   if ((i = write(to_fd, buffer, pos)) != pos) {
      fprintf(stderr, _("Cannot write data to system activity file: %s\n"), strerror(errno));
      exit(3);
   }
}


/*
 ***********************************
 * Print everything now (stats and uptime)
 ***********************************
 */
int print_stat(int curr, int write_to_file, int to_fd, int read_from_file,
	       int use_tm_start, int use_tm_end)
{
   int len, i;
   unsigned int mask;
   

   /* Get current time */
   if (!read_from_file)
      len = sprintf(stat_line, "%s  ", get_date_or_time(DT_TIME));
   else {
      if (!next_slice())
	 /* Not close enough from desired interval */
	 return 0;
      
      len = sprintf(stat_line, "%02d:%02d:%02d  ", sl_hour, sl_minute, sl_second); 
   }
   
   /* Check if time is right */
   if (use_tm_start && (strncmp(stat_line, tm_start, 8) < 0))
      /* It's too soon... */
      return 0;
   
   uptime[!curr] = uptime[curr] - uptime[!curr];

   /* Print number of processes created per second */
   if (GET_FORK(actflag))
      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((total_forks[curr] - total_forks[!curr]) * 100) / uptime[!curr],
		     (((((total_forks[curr] - total_forks[!curr]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);
   
   /* Print number of context switches per second */
   if (GET_CTXSW(actflag))
      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((context_swtch[curr] - context_swtch[!curr]) * 100) / uptime[!curr],
		     (((((context_swtch[curr] - context_swtch[!curr]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);
   
   /* Print CPU usage */
   if (GET_CPU(actflag)) {
      len += sprintf(stat_line + len, "   %3lu.%02lu%%",
		     ((cpu_user[curr] - cpu_user[!curr]) * 100) / uptime[!curr],
		     (((((cpu_user[curr] - cpu_user[!curr]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);
      
      len += sprintf(stat_line + len, "   %3lu.%02lu%%",
		     ((cpu_nice[curr] - cpu_nice[!curr]) * 100) / uptime[!curr],
		     (((((cpu_nice[curr] - cpu_nice[!curr]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);

      len += sprintf(stat_line + len, "   %3lu.%02lu%%",
		     ((cpu_system[curr] - cpu_system[!curr]) * 100) / uptime[!curr],
		     (((((cpu_system[curr] - cpu_system[!curr]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);

      len += sprintf(stat_line + len, "   %3lu.%02lu%%",
		     ((cpu_idle[curr] - cpu_idle[!curr]) * 100) / uptime[!curr],
		     (((((cpu_idle[curr] - cpu_idle[!curr]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);
   }
   
   /* Print number of interrupts per second */
   if (GET_INT(actflag)) {
      if (int_stat[NR_IRQS >> 5] & 0x80000000)
	 len += sprintf(stat_line + len, " %6lu.%02lu",
			((int_sum[curr] - int_sum[!curr]) * 100) / uptime[!curr],
			(((((int_sum[curr] - int_sum[!curr]) * 100)
			   % uptime[!curr]) * 100) / uptime[!curr]) % 100);

      for (i = 0; i < NR_IRQS; i++) {
	 mask = 1 << (i & 0x1f);
	 if (int_stat[i >> 5] & mask)
	    len += sprintf(stat_line + len, " %6lu.%02lu",
			   ((interrupts[curr][i] - interrupts[!curr][i]) * 100) / uptime[!curr],
			   (((((interrupts[curr][i] - interrupts[!curr][i]) * 100)
			      % uptime[!curr]) * 100) / uptime[!curr]) % 100);
      }
   }
   
   /* Print number of pages the system paged in and out */
   if (GET_PAGE(actflag)) {
      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((pgpgin[curr] - pgpgin[!curr]) * 100) / uptime[!curr],
		     (((((pgpgin[curr] - pgpgin[!curr]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);

      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((pgpgout[curr] - pgpgout[!curr]) * 100) / uptime[!curr],
		     (((((pgpgout[curr] - pgpgout[!curr]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);
   }
   
   /* Print number of swap pages brought in and out */
   if (GET_SWAP(actflag)) {
      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((pswpin[curr] - pswpin[!curr]) * 100) / uptime[!curr],
		     (((((pswpin[curr] - pswpin[!curr]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);

      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((pswpout[curr] - pswpout[!curr]) * 100) / uptime[!curr],
		     (((((pswpout[curr] - pswpout[!curr]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);
   }
   
   /* Print I/O stats (no distinction made between disks) */
   if (GET_IO(actflag)) {
      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((dk_drive[curr] - dk_drive[!curr]) * 100) / uptime[!curr],
		     (((((dk_drive[curr] - dk_drive[!curr]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);

      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((dk_drive_rio[curr] - dk_drive_rio[!curr]) * 100) / uptime[!curr],
		     (((((dk_drive_rio[curr] - dk_drive_rio[!curr]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);

      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((dk_drive_wio[curr] - dk_drive_wio[!curr]) * 100) / uptime[!curr],
		     (((((dk_drive_wio[curr] - dk_drive_wio[!curr]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);

      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((dk_drive_rblk[curr] - dk_drive_rblk[!curr]) * 100) / uptime[!curr],
		     (((((dk_drive_rblk[curr] - dk_drive_rblk[!curr]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);

      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((dk_drive_wblk[curr] - dk_drive_wblk[!curr]) * 100) / uptime[!curr],
		     (((((dk_drive_wblk[curr] - dk_drive_wblk[!curr]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);
   }

   printf("%s\n", stat_line);
   
   /* Write data to system activity file if necessary */
   if (write_to_file)
      write_stat(curr, to_fd);

   /* Check if time is right */
   if (use_tm_end && (strncmp(stat_line, tm_end, 8) >= 0)) {
      /* It's too late now... */
      count = 0;
      return 0;
   }
   
   return 1;
}

/*
 ***********************************
 * Print stats average
 ***********************************
 */
void print_stat_avg(int curr)
{
   int len = 10;  /* Initial length of stat_line: "Average:  " */
   int i = 0;
   unsigned int mask;

   
   strcpy(stat_line, _("Average:  "));
   
   uptime[!curr] = uptime[curr] - uptime[2];
   
   if (GET_FORK(actflag))
      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((total_forks[curr] - total_forks[2]) * 100) / uptime[!curr],
		     (((((total_forks[curr] - total_forks[2]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);
   
   if (GET_CTXSW(actflag))
      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((context_swtch[curr] - context_swtch[2]) * 100) / uptime[!curr],
		     (((((context_swtch[curr] - context_swtch[2]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);
   
   if (GET_CPU(actflag)) {
      len += sprintf(stat_line + len, "   %3lu.%02lu%%",
		     ((cpu_user[curr] - cpu_user[2]) * 100) / uptime[!curr],
		     (((((cpu_user[curr] - cpu_user[2]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);
      
      len += sprintf(stat_line + len, "   %3lu.%02lu%%",
		     ((cpu_nice[curr] - cpu_nice[2]) * 100) / uptime[!curr],
		     (((((cpu_nice[curr] - cpu_nice[2]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);

      len += sprintf(stat_line + len, "   %3lu.%02lu%%",
		     ((cpu_system[curr] - cpu_system[2]) * 100) / uptime[!curr],
		     (((((cpu_system[curr] - cpu_system[2]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);

      len += sprintf(stat_line + len, "   %3lu.%02lu%%",
		     ((cpu_idle[curr] - cpu_idle[2]) * 100) / uptime[!curr],
		     (((((cpu_idle[curr] - cpu_idle[2]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);
   }
   
   if (GET_INT(actflag)) {
      if (int_stat[NR_IRQS >> 5] & 0x80000000)
	 len += sprintf(stat_line + len, " %6lu.%02lu",
			((int_sum[curr] - int_sum[2]) * 100) / uptime[!curr],
			(((((int_sum[curr] - int_sum[2]) * 100)
			   % uptime[!curr]) * 100) / uptime[!curr]) % 100);

      for (i = 0; i < NR_IRQS; i++) {
	 mask = 1 << (i & 0x1f);
	 if (int_stat[i >> 5] & mask)
	    len += sprintf(stat_line + len, " %6lu.%02lu",
			   ((interrupts[curr][i] - interrupts[2][i]) * 100) / uptime[!curr],
			   (((((interrupts[curr][i] - interrupts[2][i]) * 100)
			      % uptime[!curr]) * 100) / uptime[!curr]) % 100);
      }
   }
   
   if (GET_PAGE(actflag)) {
      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((pgpgin[curr] - pgpgin[2]) * 100) / uptime[!curr],
		     (((((pgpgin[curr] - pgpgin[2]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);

      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((pgpgout[curr] - pgpgout[2]) * 100) / uptime[!curr],
		     (((((pgpgout[curr] - pgpgout[2]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);
   }
   
   if (GET_SWAP(actflag)) {
      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((pswpin[curr] - pswpin[2]) * 100) / uptime[!curr],
		     (((((pswpin[curr] - pswpin[2]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);

      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((pswpout[curr] - pswpout[2]) * 100) / uptime[!curr],
		     (((((pswpout[curr] - pswpout[2]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);
   }
   
   if (GET_IO(actflag)) {
      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((dk_drive[curr] - dk_drive[2]) * 100) / uptime[!curr],
		     (((((dk_drive[curr] - dk_drive[2]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);

      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((dk_drive_rio[curr] - dk_drive_rio[2]) * 100) / uptime[!curr],
		     (((((dk_drive_rio[curr] - dk_drive_rio[2]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);

      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((dk_drive_wio[curr] - dk_drive_wio[2]) * 100) / uptime[!curr],
		     (((((dk_drive_wio[curr] - dk_drive_wio[2]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);

      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((dk_drive_rblk[curr] - dk_drive_rblk[2]) * 100) / uptime[!curr],
		     (((((dk_drive_rblk[curr] - dk_drive_rblk[2]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);

      len += sprintf(stat_line + len, " %6lu.%02lu",
		     ((dk_drive_wblk[curr] - dk_drive_wblk[2]) * 100) / uptime[!curr],
		     (((((dk_drive_wblk[curr] - dk_drive_wblk[2]) * 100)
			% uptime[!curr]) * 100) / uptime[!curr]) % 100);
   }
   
   printf("\n%s\n", stat_line);
}


/*
 ***********************************
 * Main entry to the program
 ***********************************
 */
int main(int argc, char **argv)
{
   int curr = 1, opt = 1;
   unsigned int mask;
   int i, from_fd, to_fd;
   char from_file[MAX_FILE_LEN], to_file[MAX_FILE_LEN];
   int itv = 0;
   

   if (argc == 1)
      usage(argv[0]);

   from_file[0] = to_file[0] = tm_start[0] = tm_end[0] = '\0';
   init_int_stat(0);
   init_cpu_stat(0);
   
   /* Process options */
   while (opt < argc) {

      if (!strcmp(argv[opt], "-I")) {
	 if (argv[++opt]) {
	    actflag |= A_INT;
	    /*
	     * Get irq number.
	     */
	    i = atol(argv[opt++]);
	    if ((i < -1) || (i >= NR_IRQS))
	       usage(argv[0]);
	    if (i == -1) {
	       /* Save last bit, used for total number of interrupts */
	       mask = int_stat[NR_IRQS >> 5];
	       /* Set every bit */
	       init_int_stat(0xffffffff);
	       /* Restore saved bit */
	       int_stat[NR_IRQS >> 5] = mask | 0x7fffffff;
	    }
	    else
	       int_stat[i >> 5] |= 1 << (i & 0x1f);
	 }
	 else
	    usage(argv[0]);
      }
      else if (!strcmp(argv[opt], "-U")) {
	 if (argv[++opt]) {
	    actflag |= A_CPU;
	    /*
	     * Get cpu number.
	     */
	    i = atol(argv[opt++]);
	    if ((i < -1) || (i >= NR_CPUS))
	      usage(argv[0]);
	    if (i == -1) {
	       /* Save last bit, used for global cpu */
	       mask = cpu_stat[NR_CPUS >> 5];
	       /* Set every bit */
	       init_cpu_stat(0xffffffff);
	       /* Restore saved bit */
	       cpu_stat[NR_CPUS >> 5] = mask | 0x7fffffff;
	    }
	    else
	       cpu_stat[i >> 5] |= 1 << (i & 0x1f);
	 }
	 else
	    usage(argv[0]);
      }
      else if (!strcmp(argv[opt], "-o")) {
	 /* Save stats to a file */
	 if ((argv[++opt]) && strncmp(argv[opt], "-", 1) &&
	     (strspn(argv[opt], DIGITS) != strlen(argv[opt])))
	    strcpy(to_file, argv[opt++]);
	 else
	    sprintf(to_file, "%s/sa%s", SA_DIR, get_date_or_time(DT_DAY));
      }
      else if (!strcmp(argv[opt], "-f")) {
	 /* Read stats from a file */
	 if ((argv[++opt]) && strncmp(argv[opt], "-", 1) &&
	     (strspn(argv[opt], DIGITS) != strlen(argv[opt])))
	    strcpy(from_file, argv[opt++]);
	 else
	    sprintf(from_file, "%s/sa%s", SA_DIR, get_date_or_time(DT_DAY));
      }
      else if (!strcmp(argv[opt], "-s")) {
	 /* Get time start */
	 if ((argv[++opt]) && (strlen(argv[opt]) == 8))
	    strcpy(tm_start, argv[opt++]);
	 else
	    strcpy(tm_start, DEF_TMSTART);
      }
      else if (!strcmp(argv[opt], "-e")) {
	 /* Get time end */
	 if ((argv[++opt]) && (strlen(argv[opt]) == 8))
	    strcpy(tm_end, argv[opt++]);
	 else
	    strcpy(tm_end, DEF_TMEND);
      }
      else if (!strncmp(argv[opt], "-", 1) && strcmp(argv[opt], "-1")) {
	 /* Other options not previously tested */
	 for (i = 1; *(argv[opt] + i); i++) {

	    switch (*(argv[opt] + i)) {
	       
	     case 'A':
	       actflag |= A_FORK + A_PAGE + A_INT + A_IO + A_CPU + A_CTXSW + A_SWAP;
	       break;
	     case 'c':
	       actflag |= A_FORK;
	       break;
	     case 'g':
	       actflag |= A_PAGE;
	       break;
	     case 'i':
	       actflag |= A_INT;
	       int_stat[NR_IRQS >> 5] |= 0x80000000;
	       break;
	     case 'r':
	       actflag |= A_IO;
	       break;
	     case 'u':
	       actflag |= A_CPU;
	       cpu_stat[NR_CPUS >> 5] |= 0x80000000;
	       break;
	     case 'w':
	       actflag |= A_CTXSW;
	       break;
	     case 'W':
	       actflag |= A_SWAP;
	       break;
	     default:
	       usage(argv[0]);
	    }
	 }
	 opt++;
      }
      else if (!itv) { 				/* Get interval */
	 if (strspn(argv[opt], XDIGITS) != strlen(argv[opt]))
	    usage(argv[0]);
	 interval = atol(argv[opt++]);
	 if ((interval < -1) || !interval)
	   usage(argv[0]);
	 count = -1;
	 itv = 1;
      }
      else {					/* Get count value */
	 if (strspn(argv[opt], DIGITS) != strlen(argv[opt]))
	    usage(argv[0]);
	 count = atol(argv[opt++]);
	 if (count < 1)
	   usage(argv[0]);
      }
   }

   /* Default is CPU activity... */
   if (!actflag)
      actflag |= A_CPU;

   /*
    * Check option dependencies
    */
   /* Interval required if not reading from a file */
   if ((interval == -1) && !from_file[0])
      usage(argv[0]);
   /* Use time start only when reading from a file */
   if (tm_start[0] && !from_file[0]) {
      fprintf(stderr, _("Not reading from a system activity file (use -f option)\n"));
      usage(argv[0]);
   }
   /* Use time end only when reading from or wrinting to a file */
   if (tm_end[0] && !from_file[0] && !to_file[0]) {
      fprintf(stderr, _("Not using a system activity file (use -f or -o option)\n"));
      usage(argv[0]);
   }
   /* You read from a file OR you write to it... */
   if (from_file[0] && to_file[0])
      usage(argv[0]);
   
   /* Interval must be set if stat are read from or written to file */
   if (!interval && (from_file[0] || to_file[0])) {
      fprintf(stderr, _("Please give an interval value\n"));
      usage(argv[0]);
   }

   /* Open system activity files */
   open_files(from_file, to_file, &from_fd, &to_fd);
   
   init_stat_irq();
   init_stat_cpu();

   /* Print system report header */
   print_report_hdr(from_file[0], from_fd, to_file[0], to_fd);
   
   if (from_file[0]) {      /* Read stats from file */
      uptime[!curr] = 0;
      if (read_stat_from_file(!curr, from_fd)) {
	 fprintf(stderr, _("No entry found in system activity file!\n"));
	 exit(2);
      }
   }
   else {
      alarm_handler(0);     /* Set a handler for SIGALRM */
      read_uptime(!curr);
      read_stat(!curr);
   }
   
   trans_stat(2, !curr);

   if (interval) {
      if (to_file[0])
	 write_stat(!curr, to_fd);
      if (!from_file[0])
	 pause();
   }
   else {
      /* We don't use a system activity file, since interval == 0 */
      print_stat(!curr, 0, to_fd, from_file[0], 0, 0);
      exit(0);
   }
   
   /* Main loop */
   do {
      if (from_file[0]) {
	 /* Read stats from file */
	 file_uptime += file_interval * 100;
	 if (read_stat_from_file(curr, from_fd)) {
	    /* End of system activity file: display average stats now */
	    curr ^= 1; break;
	 }
	 uptime[curr] = file_uptime;
      }
      else {
	 read_uptime(curr);
	 read_stat(curr);
      }
      
      /* Print results */
      if (print_stat(curr, to_file[0], to_fd, from_file[0], tm_start[0], tm_end[0])
	  && (count > 0))
	 count--;
      fflush(stdout);
      
      if (count != 0) {
	 if (!from_file[0])
	    pause();
	 curr ^= 1;
      }
   }
   while (count != 0);
   
   /* Close system activity files */
   close_files(from_file, to_file, from_fd, to_fd);

   /* Print stat average */
   print_stat_avg(curr);
   
   return 0;
}

